/*-----------------------------------------------------------------------

File  : ccl_rewrite.c

Author: Stephan Schulz

Contents
 
  Rewriting. For all functions, if index != NULL, delete rewritten
  terms and superterms from the index!

Changes

<1> Tue May 26 19:47:52 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_rewrite.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

long RewriteAttempts = 0;
long RewriteSucesses = 0;

#ifndef NDEBUG
bool RewriteWarnLeftSide = false;
#endif

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*-----------------------------------------------------------------------
//
// Function: instance_is_rule()
//
//   Return true if lside->rside is a rule, i.e. lside>rside (for the
//   instantiated terms) and rside contains no unbound
//   variables. Assumes that uninstantated terms lside and rside are
//   uncomparable! 
//
// Global Variables: -
//
// Side Effects    : TOGreater() can manipulate TPOpFlag of terms.
//
/----------------------------------------------------------------------*/

static bool instance_is_rule(OCB_p ocb, Term_p lside, Term_p rside,
			     Subst_p subst)

{   
   if(TermHasUnboundVariables(rside))
   {
      return false;
   }
   if(SubstIsRenaming(subst)) /* Save comparisons */
   {
      return false;
   }
   return TOGreater(ocb, lside, rside, DEREF_ONCE, DEREF_ONCE);
}



/*-----------------------------------------------------------------------
//
// Function: term_is_top_rewritable()
//
//   Return true if the term is rewritable with new_demod at the top
//   position, false otherwise.
//
// Global Variables: -
//
// Side Effects    : Temporary variable bindings (backtracked)
//
/----------------------------------------------------------------------*/

static bool term_is_top_rewritable(OCB_p ocb, Term_p term, Clause_p
				   new_demod)
{
   Subst_p subst = SubstAlloc();
   Eqn_p   eqn;
   bool    res = false;
   
   assert(new_demod->pos_lit_no == 1);
   assert(new_demod->neg_lit_no == 0);
   assert(!TermIsVar(term));
   
   eqn = new_demod->literals;

   DEBUGMARK(RW_MATCH_WATCH, "Matching ");
   DEBUG(RW_MATCH_WATCH, 
	 ClausePrint(stdout, new_demod, true););
   DEBUGOUT(RW_MATCH_WATCH, " (l) onto ");
   DEBUG(RW_MATCH_WATCH, 
	 TermPrint(stdout, term, ocb->sig, DEREF_NEVER););
   DEBUGOUT(RW_MATCH_WATCH, "\nInstantiated lside: ");
   DEBUG(RW_MATCH_WATCH,
	 TermPrint(stdout, eqn->lterm, ocb->sig, DEREF_ONCE););
   DEBUGOUT(RW_MATCH_WATCH, "\n");
   if(SubstComputeMatch(eqn->lterm, term, subst, TBTermEqual))
   {
      if((EqnIsOriented(eqn) 
	  || instance_is_rule(ocb, eqn->lterm, eqn->rterm, subst))
	 &&
	 (!TermCellQueryProp(term, TPRestricted) ||
	  !SubstIsRenaming(subst)))
      {
	 res = true;
      }
      SubstBacktrack(subst);
   }
   if(!res && !EqnIsOriented(eqn))
   {
      if(SubstComputeMatch(eqn->rterm, term, subst, TBTermEqual))
      {
	 if(instance_is_rule(ocb, eqn->rterm, eqn->lterm, subst))
	    /* If instance is rule -> subst is no renaming! */
	 {
	    res = true;
	 }
      }
   }
   SubstDelete(subst);
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: term_is_rewritable()
//
//   Return true if the term is rewritable with new_demod, false
//   otherwise. Set nf_date[0,1] on non-rewritable terms to nf_date
//   (i.e. assumes thate term is in normal for with respect to earlier
//   systems).
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms.
//
/----------------------------------------------------------------------*/

static bool term_is_rewritable(OCB_p ocb, Term_p term, Clause_p
			       new_demod, SysDate nf_date)
{
   int i;
   bool res = false;

   /* printf("term_is_rewritable()...\n"); */
   
   if(TermIsVar(term))
   {
      return false;
   }
   if(TermCellQueryProp(term, TPIsRewritable))
   {
      return true;
   }
   if(SysDateCompare(term->nf_date[RewriteAdr(FullRewrite)], nf_date) == DateEqual)
   {
      return false;
   }
   for(i=0; i<term->arity; i++)
   {
      if(term_is_rewritable(ocb, term->args[i], new_demod, nf_date))
      {
	 res = true;
	 break;
      }
   }
   if(res)
   {
      TermCellSetProp(term, TPIsRewritable);
      return true;
   }
   if(term_is_top_rewritable(ocb, term, new_demod))
   {
      TermCellSetProp(term, TPIsRewritable);
      return true;
   }
   term->nf_date[RewriteAdr(RuleRewrite)] = term->nf_date[RewriteAdr(FullRewrite)] = nf_date;
   
   /* printf("...term_is_rewritable() = false (no match)\n"); */
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: eqn_has_rw_max_side()
//
//   Return true if the equation has a rewritable maximal term. Try
//   both terms to determine irreducible terms.
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms.
//
/----------------------------------------------------------------------*/

static bool eqn_has_rw_max_side(OCB_p ocb, Eqn_p eqn, Clause_p
				new_demod, SysDate nf_date)
{
   bool resl, resr;

   resl = term_is_rewritable(ocb, eqn->lterm, new_demod, nf_date);
   resr = term_is_rewritable(ocb, eqn->rterm, new_demod, nf_date);

   if(EqnIsOriented(eqn))
   {
      return resl;
   }
   return resl|resr;
}
   
/*-----------------------------------------------------------------------
//
// Function: clause_has_rw_max_side()
//
//   Return true if the clause has a rewritable maximal literal.
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms.
//
/----------------------------------------------------------------------*/

static bool clause_has_rw_max_side(OCB_p ocb, Clause_p clause,
				   Clause_p new_demod, SysDate
				   nf_date)
{
   Eqn_p handle;
   
   for(handle = clause->literals; handle; handle = handle->next)
   {
      if(EqnIsMaximal(handle) && eqn_has_rw_max_side(ocb, handle,
						     new_demod, nf_date))
      {
	 return true;
      }	 
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: find_clauses_with_rw_max_sides()
//
//   A non-index-using implementation of
//   FindClausesWithRewritableMaxSides(). See there...
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms.
//
/----------------------------------------------------------------------*/

static void find_clauses_with_rw_max_sides(OCB_p ocb, ClauseSet_p set,
					   PTree_p* results, Clause_p
					   new_demod, SysDate nf_date)
{
   Clause_p handle;
   
   assert(new_demod->pos_lit_no == 1);
   assert(new_demod->neg_lit_no == 0);
   
   for(handle = set->anchor->succ; handle != set->anchor; handle =
	  handle->succ)
   {
      if(clause_has_rw_max_side(ocb, handle, new_demod, nf_date))
      {
	 PTreeStore(results, handle);
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: get_max_demod_date()
//
//   Return the oldest date of a set of demodulators in the NULL
//   terminated array demodulators.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

SysDate get_max_demod_date(ClauseSet_p *demodulators)
{
   int i;
   SysDate res = SysDateCreationTime();
   
   for(i=0; demodulators[i]; i++)
   {
      res = SysDateMaximum(res, demodulators[i]->date);
   }
   return res;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: TermFindRewriteClause()
//
//   Given a term and a (unit) clause set, find a unit clause that is
//   able to rewrite the term. Set pos accordingly, return matching
//   clause. The matching term is instantiated, the match is stored in
//   subst. Clauses are tried in the order defined by
//   demodulators->eval_indices[0], that is, the set must have at
//   least one evaluations (probably AddWeight()).
//
//   This function will probably be replaced by an tree indexing based
//   function.
//
// Global Variables: -
//
// Side Effects    : Instantiates, changes pos
//
/----------------------------------------------------------------------*/

Clause_p TermFindRewriteClause(OCB_p ocb, Term_p term, ClauseSet_p
			       demodulators, ClausePos_p pos, Subst_p
			       subst)
{
   Clause_p      handle, res = NULL;
   Eval_p        key;
   PStack_p      stack;
   PStackPointer state = PStackGetSP(subst);
   Eqn_p         eqn;

   DEBUGMARK(RW_INTERFACE_WATCH2, "TermFindRewriteClause()...\n");   
   assert(term);
   assert(demodulators);
   assert(pos);
   assert(demodulators->eval_indices[0]
	  ||(demodulators->anchor->succ==demodulators->anchor));
   
   RewriteAttempts++;   
   stack = EvalTreeTraverseInit(demodulators->eval_indices[0]);
   
   while((key = EvalTreeTraverseNext(stack)))
   {
      handle = key->object;

      assert(handle->pos_lit_no == 1);
      assert(handle->neg_lit_no == 0);
      
      if((EqnIsOriented(handle->literals) && 
	  SysDateCompare(term->nf_date[0], handle->date)!=DateEarlier)
	 ||
	 (!EqnIsOriented(handle->literals) && 
	  SysDateCompare(term->nf_date[1], handle->date)!=DateEarlier))
      {
	 DEBUGMARK(RW_INTERFACE_WATCH2,
		   "...TermFindRewriteClause() (Used dates)\n");  
	 break;
      }
      
      pos->clause  = handle;
      eqn = handle->literals;
      
      DEBUGMARK(RW_MATCH_WATCH, "Matching ");
      DEBUG(RW_MATCH_WATCH, 
	    ClausePrint(stdout, pos->clause, true););
      DEBUGOUT(RW_MATCH_WATCH, " (l) onto ");
      DEBUG(RW_MATCH_WATCH, 
	    TermPrint(stdout, term, ocb->sig, DEREF_NEVER););
      DEBUGOUT(RW_MATCH_WATCH, "\nInstantiated lside: ");
      DEBUG(RW_MATCH_WATCH, 
	    TermPrint(stdout, eqn->lterm, ocb->sig, DEREF_ONCE););
      DEBUGOUT(RW_MATCH_WATCH, "\n");
      
      if(SubstComputeMatch(eqn->lterm, term, subst, TBTermEqual))
      {
	 if((EqnIsOriented(eqn) 
	     || instance_is_rule(ocb, eqn->lterm, eqn->rterm, subst))
	    &&
	    (!TermCellQueryProp(term, TPRestricted) ||
	     !SubstIsRenaming(subst)))
	 {
	    res = handle;
	    pos->literal = eqn;
	       pos->side = LeftSide;
	       break;
	 }
	 SubstBacktrackToPos(subst, state);
      }

      if(!EqnIsOriented(eqn))
      {
	 DEBUGMARK(RW_MATCH_WATCH, "Matching ");
	 DEBUG(RW_MATCH_WATCH, 
	       ClausePrint(stdout, pos->clause, true););
	 DEBUGOUT(RW_MATCH_WATCH, " (r) onto ");
	 DEBUG(RW_MATCH_WATCH, 
	       TermPrint(stdout, term, ocb->sig, DEREF_NEVER););
	 DEBUGOUT(RW_MATCH_WATCH, "\n");
	 if(SubstComputeMatch(eqn->rterm, term, subst, TBTermEqual))
	 {
	    if(instance_is_rule(ocb, eqn->rterm, eqn->lterm, subst)
	       &&
	       (!TermCellQueryProp(term, TPRestricted) ||
		!SubstIsRenaming(subst)))
	    {
	       res = handle;
	       pos->literal = eqn;
	       pos->side = RightSide;
	       break;
	    }	    
	    SubstBacktrackToPos(subst, state);
	 }	 
      }
   }
   EvalTreeTraverseExit(stack);

   DEBUGMARK(RW_INTERFACE_WATCH2, "...TermFindClause()\n");   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TermIndexedFindDemodulator()
//
//   Find a demodulator via demodulators->demod_index. This is the
//   replacement function for TermFindRewriteClause()
//   above. Unfortunatly, it has a different calling interface - which
//   should be much more efficent. TermRewriteWithClauseSet() will
//   hopefully take care of the resulting mess transparently.
//
// Global Variables: -
//
// Side Effects    : Instantiates
//
/----------------------------------------------------------------------*/

ClausePos_p TermIndexedFindDemodulator(OCB_p ocb, Term_p term,
				       SysDate date,
				       ClauseSet_p demodulators,
				       Subst_p subst)
{
   Eqn_p       eqn;   
   PDSearch_p  state;  
   ClausePos_p pos, res = NULL;
   
   DEBUGMARK(RW_INTERFACE_WATCH2,
	     "TermIndexedFindDemodulator()...\n");
   
   assert(term);
   assert(demodulators);
   assert(demodulators->demod_index);
   assert(term->weight == TermWeight(term, DEFAULT_VWEIGHT,
				     DEFAULT_FWEIGHT));

   RewriteAttempts++;   
   
   DEBUGMARK(RW_MATCH_WATCH, "\nSearching match for: ");
   DEBUG(RW_MATCH_WATCH, TermPrint(stdout, term, ocb->sig,
				   DEREF_NEVER);
	 printf("\n"););
   
   state = PDSearchAlloc(demodulators->demod_index, term, date);

   while((pos = PDTreeFindNextDemodulator(state, subst)))
   {
      eqn = pos->literal;
      
      if((EqnIsOriented(eqn)&&
	  SysDateCompare(term->nf_date[RewriteAdr(RuleRewrite)],
			 pos->clause->date)!=DateEarlier)
	 ||
	 (!EqnIsOriented(eqn)&&
	  SysDateCompare(term->nf_date[RewriteAdr(FullRewrite)],
			 pos->clause->date)!=DateEarlier))
      {
	 continue;
      }

      switch(pos->side)
      {
      case LeftSide:
	    DEBUGMARK(RW_MATCH_WATCH, "Found: ");
	    DEBUG(RW_MATCH_WATCH, 
		  ClausePrint(stdout, pos->clause, true);
		  printf(" (l), instantiated: ");
		  TermPrint(stdout, pos->clause->literals->lterm,
			    ocb->sig, DEREF_ONCE);
		  printf(" <=> ");
		  TermPrint(stdout, pos->clause->literals->rterm,
			    ocb->sig, DEREF_ONCE);
		  printf("\n");
	       );
	    if((EqnIsOriented(eqn) 
		|| instance_is_rule(ocb, eqn->lterm, eqn->rterm, subst))
	       &&
	       (!TermCellQueryProp(term, TPRestricted) ||
		!SubstIsRenaming(subst)))
	    {
	       res = pos;
	    }
	    break;
      case RightSide:
	    DEBUGMARK(RW_MATCH_WATCH, "Found: ");
	    DEBUG(RW_MATCH_WATCH, 
		  ClausePrint(stdout, pos->clause, true);
		  printf(" (r), instantiated: ");
		  TermPrint(stdout, pos->clause->literals->lterm,
			    ocb->sig, DEREF_ONCE);
		  printf(" <=> ");
		  TermPrint(stdout, pos->clause->literals->rterm,
			    ocb->sig, DEREF_ONCE);
		  printf("\n");
	       );
	    assert(!EqnIsOriented(eqn));
	    if(instance_is_rule(ocb, eqn->rterm, eqn->lterm, subst)
	       &&
	       !TermCellQueryProp(term, TPRestricted)) 
	       /* Case SubstIsRenaming(subst) already eleminated in
		  instance_is_rule! */
	    {
	       res = pos;
	    }
	    break;
      default: 
	    assert(false);
	    break;
      }
      if(res)
      {
	 break;
      }
   }
   PDSearchFree(state);

   DEBUGMARK(RW_INTERFACE_WATCH2,
	     "...TermIndexedFindDemodulator()\n");
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TermRewriteWithClauseSet()
//
//   Rewrite the given term at root position with the first matching,
//   orientable rule from demodulators. Return true if a matching rule
//   was found, false otherwise.
//
// Global Variables: -
//
// Side Effects    : Changes term and termbank, by TBTermReplace()
//
/----------------------------------------------------------------------*/

bool TermRewriteWithClauseSet(OCB_p ocb, TB_p bank, Term_p term,
			      SysDate date, ClauseSet_p demodulators)
{
   bool        res   = false;
   Subst_p     subst = SubstAlloc();
   ClausePos_p pos   = NULL;
   Clause_p    demod;
   Term_p      repl;
   void*       found = NULL;
   
   assert(!TermCellQueryProp(term, TPIsIndexed));
   DEBUGMARK(RW_INTERFACE_WATCH2, "TermRewriteWithClauseSet()...\n");   

   if(demodulators->demod_index)
   {
      pos = TermIndexedFindDemodulator(ocb, term, date, demodulators,
				       subst);
      found = pos;
   }
   else
   {
      pos = ClausePosAlloc();
      demod = TermFindRewriteClause(ocb, term, demodulators, pos,
				    subst);
      found = demod;
   }
   if(found)
   {
      DEBUGMARK(RW_REWRITE_WATCH, "Rewriting ");
      DEBUG(RW_REWRITE_WATCH,
	    TermPrint(stdout, term, bank->sig, DEREF_NEVER););
      DEBUGOUT(RW_REWRITE_WATCH, " with ");
      DEBUG(RW_REWRITE_WATCH,
	    ClausePrint(stdout, pos->clause, true);
	    printf("(%c)\n", pos->side == LeftSide? 'l':'r'););
      
      RewriteSucesses++;

      repl = ClausePosGetOtherSide(pos);
      repl = TBInsertInstantiated(bank, repl);
      TBTermReplace(bank, term, repl);
      
      DEBUGMARK(RW_REWRITE_WATCH, "Result ");
      DEBUG(RW_REWRITE_WATCH,
	    TermPrint(stdout, repl, bank->sig, DEREF_NEVER););
      DEBUGOUT(RW_REWRITE_WATCH, "\n");
      
      res = true;
   }
   SubstDelete(subst);
   
   if(!demodulators->demod_index)
   {
      ClausePosFree(pos);
   }

   DEBUGMARK(RW_INTERFACE_WATCH2, "...TermRewriteWithClauseSet()\n");   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TermRewriteWithClauseSetList()
//
//   Rewrite a term with the sets of demodulators.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

bool TermRewriteWithClauseSetList(OCB_p ocb, TB_p bank, Term_p term,
				  ClauseSet_p* demodulators,
				  RewriteLevel level)
{
   int  i;
   bool res = false;

   assert(level);

   for(i=0; i<level; i++)
   {
      assert(demodulators[i]);
      if(SysDateCompare(term->nf_date[level-1], demodulators[i]->date) ==
	 DateEarlier)
      {
	 res = TermRewriteWithClauseSet(ocb, bank, term,
					term->nf_date[level-1],
					demodulators[i]);
	 if(res)
	 {
	    break;
	 }
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TermComputeLINormalform()
//
//   Compute a leftmost-innermost normal form of term. This is more
//   tricky than expected, as supertems on the stack may change. I'll
//   start with a recursive version and see how it works. This uses
//   dates to minimize rewrite-attempts: If the normal form of the
//   term is younger than the clause sets, no further rewrite-attempt
//   on this term is made.
//
// Global Variables: -
//
// Side Effects    : Changes terms, updates nf_date.
//
/----------------------------------------------------------------------*/

bool TermComputeLINormalform(OCB_p ocb, TB_p bank, Term_p term,
			     ClauseSet_p *demodulators, RewriteLevel
			     level)
{
   bool    modified = true, res = false;
   SysDate demod_date = ClauseSetListGetMaxDate(demodulators, level);

   DEBUGMARK(RW_INTERFACE_WATCH2, "TermComputeLINormalform()...\n");   

   if(level == NoRewrite)
   {
      return false;
   }
   if(SysDateCompare(term->nf_date[level-1], demod_date)!=DateEarlier)
   {
      DEBUGMARK(RW_INTERFACE_WATCH2, "...TermComputeLINormalform() (Used dates)\n");  
      assert(term->weight == TermWeight(term, DEFAULT_VWEIGHT,
					DEFAULT_FWEIGHT));
      return false;
   }      
   if(TermIsVar(term))
   {
      assert(term->weight == DEFAULT_VWEIGHT);
      return false;
   }
   while(modified)
   {
      int i;
            
      TermGetRef(&term, term);
      for(i=0; i<term->arity; i++)
      {
	 modified = TermComputeLINormalform(ocb, bank,
					    term->args[i],
					    demodulators, level);
	 res = res || modified ;
      }
      if(!TermIsVar(term)) /* Remember, term itself may have been
			      replaced */
      {
	 term->weight = DEFAULT_FWEIGHT;
	 for(i=0; i<term->arity; i++)
	 {
	    term->weight += term->args[i]->weight;
	 }
      }	
      assert(term->weight == TermWeight(term, DEFAULT_VWEIGHT, DEFAULT_FWEIGHT));
      modified = TermRewriteWithClauseSetList(ocb, bank, term,
					      demodulators, level);
      TermReleaseRef(&term);
      res = res || modified;
   }
   term->nf_date[RewriteAdr(RuleRewrite)] = demod_date;
   if(level == FullRewrite)
   {
      term->nf_date[RewriteAdr(FullRewrite)] = demod_date;
   }
   TermCellDelProp(term, TPIsRewritable);
   
   DEBUGMARK(RW_INTERFACE_WATCH2, "...TermComputeLINormalform()\n");   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnComputeLINormalform()
//
//   Compute the normal form of maximal, minimal or both terms in an
//   equation. Return true if a term was rewritten.
//
// Global Variables: -
//
// Side Effects    : Changes terms and term bank
//
/----------------------------------------------------------------------*/

bool EqnComputeLINormalform(OCB_p ocb, TB_p bank, Eqn_p eqn,
			    ClauseSet_p* demodulators,
			    RewriteLevel level)
{   
   bool res1 = false, res2 = false;

   DEBUGMARK(RW_INTERFACE_WATCH1, "EqnComputeLINormalform()...\n");
   res1 = TermComputeLINormalform(ocb, bank, eqn->lterm,
				  demodulators, level);
   res2 = TermComputeLINormalform(ocb, bank, eqn->rterm,
				  demodulators, level);
   DEBUGMARK(RW_INTERFACE_WATCH1, "EqnComputeLINormalform()...\n");
   return res1 || res2;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseComputeLINormalform()
//
//   Compute the normal form of terms in a clause. Return true if a
//   term was rewritten. 
//
// Global Variables: -
//
// Side Effects    : Changes terms and term bank
//
/----------------------------------------------------------------------*/

bool ClauseComputeLINormalform(OCB_p ocb, TB_p bank, Clause_p clause,
			       ClauseSet_p *demodulators,
			       RewriteLevel level)
{
   Eqn_p handle;
   bool tmp, res=false;

   DEBUGMARK(RW_INTERFACE_WATCH1, "ClauseComputeLINormalform()...\n");
   DEBUG(RW_INTERFACE_WATCH1, printf("# Normalizing ");
	 ClausePrint(stdout, clause, true);printf("\n"););

   for(handle = clause->literals; handle; handle=handle->next)
   {
      tmp = EqnComputeLINormalform(ocb, bank, handle, demodulators,
				   level);
      res = res || tmp;
   }
   DEBUGMARK(RW_INTERFACE_WATCH1, "...ClauseComputeLINormalform()\n");   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetComputeLINormalform()
//
//   Compute a normal form for terms in all clauses in set with
//   respect to clauses in demodulators up to level. Returns true if a
//   term was rewritten.
//
// Global Variables: -
//
// Side Effects    : Changes clauses
//
/----------------------------------------------------------------------*/

bool ClauseSetComputeLINormalform(OCB_p ocb, TB_p bank, ClauseSet_p
				  set, ClauseSet_p *demodulators,
				  RewriteLevel level)
{
   Clause_p handle;
   bool     res = false, tmp;

   DEBUGMARK(RW_INTERFACE_WATCH1,
	     "ClauseSetComputeLINormalform()...\n");

   for(handle=set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      tmp = ClauseComputeLINormalform(ocb, bank, handle,
				      demodulators, level);
      res = res || tmp;
   }
   DEBUGMARK(RW_INTERFACE_WATCH1, "...ClauseSetComputeLINormalform()\n");
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: FindClausesWithRewritableMaxSides()
//
//   Add pointers to all clauses in set for which (potentially)
//   maximal sides can be rewritten with new_demod to the results
//   tree. This uses two optimizations: 
//
//   1) If a terms nf_date is equal to nf_date, the term (and all
//      subterms) are considered to be irreducible with respect to
//      new_demod (hence nf_date should be bigger than any nf_date of
//      terms in clauses). If a term and all its subterms are
//      irreducible by demod, it's date will be set to nf_date.
//   2) If a terms nf_date is equal to SysDateCreationTime(), the term
//      is considered to be reducible with respect to new_demod. If a
//      term is rewritable, it's nf_date will be set to
//      SysDateCreationTime(). 
//
// Global Variables: -
//
// Side Effects    : Changes nf_dates of terms
//
/----------------------------------------------------------------------*/

void FindClausesWithRewritableMaxSides(OCB_p ocb, ClauseSet_p set,
				       PTree_p* results, Clause_p
				       new_demod, SysDate nf_date)
{   
   find_clauses_with_rw_max_sides(ocb, set, results, new_demod,
				  nf_date);
   /* Later: Use the index if it exists */
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


