/*-----------------------------------------------------------------------

File  : cto_proofcontrol.c

Author: Stephan Schulz

Contents
 
  Basic functions for proof control objects.

Changes

<1> Fri Oct 16 16:20:00 MET DST 1998
    Salvaged from cco_proofstate.h, forked control and state.

-----------------------------------------------------------------------*/

#include "cco_proofcontrol.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ProofControlAlloc()
//
//   Return an initialized proof control object.
//
// Global Variables: -
//
// Side Effects    : memory operations
//
/----------------------------------------------------------------------*/

ProofState_p ProofControlAlloc(ProofState_p init_state, TermOrdering
			       ordertype, TOWeightGenMethod
			       to_weight_gen, TOPrecGenMethod
			       to_prec_gen, long to_const_weight,
			       WFCB_p *weight_functions, long
			       *weight_distrib, RewriteLevel
			       forward_demod)
{
   ProofControl_p handle = ProofControlCellAlloc();

   handle->ordering            = TOCreateOrdering(state, ordertype,
						  to_weight_gen,
						  to_prec_gen,
						  to_const_weight);
   handle->heuristic           = HCBCreateHeuristic(state,
						    weight_functions,
						    weight_distrib);
   handle->forward_demod       = forward_demod;

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ProofControlFree()
//
//   Free a ProofControlCell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ProofControlFree(ProofControl_p junk)
{
   assert(junk);
   if(junk->ordering)
   {
      OCBFree(junk->ordering);
   }
   if(junk->heuristic)
   {
      HCBFree(junk->heuristic);
   }


   ProofControlCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: TOCreateOrdering()
//
//   Given a proof state, generate a (hopefully suitable) ordering for
//   it. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

OCB_p TOCreateOrdering(ProofState_p state, TermOrdering ordertype,
		       TOWeightGenMethod to_weight_gen,
		       TOPrecGenMethod to_prec_gen, long
		       to_const_weight)
{
   OCB_p handle;
   
   switch(ordertype)
   {      
   case LPO:
	 handle = OCBAlloc(LPO, state->signature);
	 TOGeneratePrecedence(handle, state->signature, state->axioms,
			      to_prec_gen);
	 /* LPOGenPrec(handle, state->signature, state->axioms); */
	 OUTPRINT(1, "# Creating term ordering: LPO selected\n");
	 break;
   case KBO:
	 handle = OCBAlloc(KBO, state->signature);
	 TOGeneratPrecedence(handle, state->signature, state->axioms,
			     to_prec_gen);
	 TOGenerateWeights(handle, state->axioms,
			   state->to_weight_gen,
			   state->to_const_weight);
	 OUTPRINT(1, "# Creating term ordering: KBO selected\n");
	 break;
   case NoOrdering:
	 /* Do some magic...currently default to KBO2! -> Fall through! */
   case KBO2:
	 handle = OCBAlloc(KBO2, state->signature);
	 TOGeneratePrecedence(handle, state->signature, state->axioms,
			      to_prec_gen);
	 TOGenerateWeights(handle, state->axioms,
			   state->to_weight_gen,
			   state->to_const_weight);
	 OUTPRINT(1, "# Creating term ordering: KBO2 selected\n");
	 break;
   case RPO:
	 assert(false && "RPO not yet implemented!");
	 handle = NULL;
	 break;
   default:
	 assert(false);
	 handle = NULL;
	 break;
   }
   return handle;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
