/*-----------------------------------------------------------------------

File  : cte_termtypes.h

Author: Stephan Schulz

Contents
 
  Declarations for the basic term type and primitive functions, mainly
  on single term cells. This module mostly provides only
  infrastructure for higher level modules.

Changes

<1> Tue Feb 24 01:23:24 MET 1998
    Ripped out of the now obsolete cte_terms.h

-----------------------------------------------------------------------*/

#ifndef CTE_TERMTYPES

#define CTE_TERMTYPES

#include <cte_signature.h>
#include <clb_sysdate.h>
#include <clb_ptrees.h>
#include <clb_properties.h>



/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/

#define DEFAULT_VWEIGHT  1  /* This has to be an integer > 0! */
#define DEFAULT_FWEIGHT  2  /* This has to be >= DEFAULT_VWEIGHT */

typedef enum 
{
   TPIgnoreProps  = 0,  /* For masking properties out */
   TPRestricted   = 1,  /* Rewriting is restricted on this term */
   TPTopPos       = 2,  /* This cell is a entry point */
   TPBasicPos     = 4,  /* Basic position on this term */
   TPPredPos      = 8,  /* This is an previous predicate position
                           morphed into a term */
   TPIsRewritable = 16, /* Term is known to be rewritable with respect
			   to a current rule or rule set. Used for
			   removing backward-rewritable
			   clauses. Absence of this flag does not mean
			   that the term is in any kind of normal
			   form! */
   TPIsIndexed    = 32, /* Term is in an index and has to be removed if
			  rewritten */
   TPSpecialFlag  = 128,/* For internal use */
   TPOpFlag       = 256,/* For internal use */
   TPOutputFlag   = 512 /* Has this term already been printed? */
}TermProperties;

typedef enum  /* See CLAUSES/ccl_rewrite.c for more */
{
   NoRewrite = 0,     /* Just for completness */
   RuleRewrite = 1,   /* Rewrite with rules only */
   FullRewrite = 2    /* Rewrite with rules and equations */
}RewriteLevel;


typedef struct termcell
{
   TermProperties   properties;    /* Like basic, lhs, top */
   FunCode          f_code;        /* Top symbol of term */
   int              arity;         /* Redundant, but saves handing
				      around the signature all the
				      time */
   struct termcell* *args;         /* Pointer to array of arguments */
   struct termcell* binding;       /* For variable bindings,
				      potentially for temporary
				      rewrites - it might be possible
				      to combine the previous two in a
				      union. */
   unsigned long    entry_no;      /* Counter for terms in a given
                                      termbank - needed for
                                      administration and external
                                      representation */
   SysDate          nf_date[FullRewrite];
                                   /* Term is in normal form with
				      respect to the demod-set at
				      date nf_date... */
   long             weight;        /* Weight of the term, computed
				      during normal form
				      building. Should be valid
				      whenever nf_date[whatever]!=
				      SysDateCreationTime().*/
   long             ref_count;     /* How often is this term
				      referenced? If this value is -1,
				      the term is unshared (even if
				      the ref_count is 0, the term can
				      stil be stored in a term bank!)
				      Note that variables in unshared
				      terms are shared (but are
				      administered in separate
				      variable banks)! */
   PTree_p          superterms;    /* Superterms, taking care of
				      "internal" references,
				      i.e. references resulting from
				      subterm relations in the term
				      bank */ 
   PTree_p          ext_refs;      /* External references,
				      i.e. positions of term pointers
				      that have to be updated if the
				      term is replaced */
   int              balance;       /* AVL stuff */
   int              passed;
   struct termcell* lson;
   struct termcell* rson;
}TermCell, *Term_p, **TermRef;


typedef long DerefType, *DerefType_p;

#define DEREF_ALWAYS -1
#define DEREF_NEVER   0
#define DEREF_ONCE    1



/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/

/* Functions which take two terms and return a boolean, i.e. test for
   equality */

typedef bool (*TermEqualTestFun)(Term_p t1, Term_p t2);

#define TERMS_INITIAL_ARGS 10

#define RewriteAdr(level) (assert(level),(level)-1)
#define TermIsVar(t) ((t)->f_code < 0)
#define TermIsConst(t)(!TermIsVar(t) && ((t)->arity==0))
#define TermIsShared(t) ((t)->ref_count != -1-)

#define TermCellSetProp(term, prop) SetProp((term), (prop))
#define TermCellDelProp(term, prop) DelProp((term), (prop))

/* Are _all_ properties in prop set in term? */
#define TermCellQueryProp(term, prop) QueryProp((term), (prop))

/* Are any properties in prop set in term? */
#define TermCellIsAnyPropSet(term, prop) IsAnyPropSet((term), (prop))

#define TermCellAlloc() (TermCell*)SizeMalloc(sizeof(TermCell))
#define TermCellFree(junk)         SizeFree(junk, sizeof(TermCell))
#define TermArgArrayAlloc(arity) ((Term_p*)SizeMalloc((arity)*sizeof(Term_p)))
#define TermArgArrayFree(junk, arity) SizeFree((junk),(arity)*sizeof(Term_p))
Term_p  DefaultTermCellAlloc();
void    TermFree(Term_p junk);

void    TermSetProp(Term_p term, DerefType deref, TermProperties prop);
void    TermDelProp(Term_p term, DerefType deref, TermProperties prop);

Term_p  TermDeref(Term_p term, DerefType_p deref);

void    TermTopFree(Term_p junk); 

Term_p* TermArgListCopy(Term_p source);
Term_p  TermTopCopy(Term_p source);

#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





